// "How to make the GBA lag"-101
// /Mic, 2000 | stabmaster_@hotmail.com | come.to/progzone
//
//
// How to compile:
//  zarmasm -CPU ARM7TDMI -Littleend start.asm
//  zarmasm -CPU ARM7TDMI -Littleend lag_data.asm
//  zarmcc -c -Wall -Otime -ansic -fpu none -Littleend -cpu ARM7TDMI -apcs /narrow/noswst lag.c -o lag.o
//  zarmlink -bin -first start.o start.o lag.o lag_data.o -map -ro-base 0x08000000 -o lag.bin
//
//
// I should probably use fixed-point instead of floating-point but I was too
// lazy to fix that.
//


#include <stdlib.h>
#include <math.h>
#include "gbal.h"


extern u8 lbl_backdrop;
extern u8 lbl_font;

long camX,camY,camZ;

u8 *backdrop=&lbl_backdrop;
u8 *font=&lbl_font;
int font_width, font_height, char_size;

POINT triangle[3];
float points[9],
      backup[9] = {0.0f,-12.0f,0.0f, -12.0f,12.0f,0.0f, 12.0f,12.0f,0.0f};
float sine[256], cosine[256];
char message[9] = "Lag / Mic";



void Rotate(float *src, float *dest, int rx, int ry, int rz, int num_points)
{
int i, offs;
float xsa,xca, ysa,yca, zsa,zca;
float xr,yr,zr,f1;

 xsa = sine[rx&255];
 xca = cosine[rx&255];
 ysa = sine[ry&255];
 yca = cosine[ry&255];
 zsa = sine[rz&255];
 zca = cosine[rz&255];

 offs = 0;
 for (i=0; i<num_points; i++)
 {
  xr = src[offs];
  yr = src[offs+1];
  zr = src[offs+2];
    f1 = yr;
    yr = ((xca*yr)-(xsa*zr));
    zr = ((xsa*f1)+(xca*zr));
    f1 = xr;
    xr = ((yca*xr)-(ysa*zr));
    zr = ((ysa*f1)+(yca*zr));
    f1 = xr;
    xr = ((zca*xr)-(zsa*yr));
    yr = ((zsa*f1)+(zca*yr));
  dest[offs] = xr;
  dest[offs+1] = yr;
  dest[offs+2] = zr;
  offs += 3;
 }
}



void Project(float *src,long *dest,int num_points)
{
int i,si,di;
float sx,sy,sz;
  si = 0;
  di = 0;
  for (i=0; i<num_points; i++)
  {
    sx = src[si]; 
    sy = src[si+1];
    sz = src[si+2];
    si += 3;
    sz += (float)(camZ);
    if (sz!=0)
    {
      dest[di] = (long)((FOCAL*sx)/sz)+camX;
      dest[di+1] = (long)((FOCAL*sy)/sz)+camY;
    }else
    {
      dest[di] = camX;
      dest[di+1] = camY;
    }
    di += 5;
  }
}




void mirror()
{
u8 *src;
 src = (u8 *)(0x06005CD0);
 __asm
 {
  mov r0,src
  mov r1,#0x06005DC0
  mov r2,#60
  mov r4,#0
  mov r6,#43690  
  mYloop:
    mov r3,240
    mXloop:
      ldrb r5,[r0]
      add r5,r5,#64  
      strb r5,[r1]
      add r0,r0,#1
      add r1,r1,#1
      subs r3,r3,#1
    bne mXloop
    mov r0,src
    sub r0,r0,#240
    adds r4,r4,r6
    cmp r4,#65535
    subgt r0,r0,#240
    and r4,r4,#65535
    mov src,r0
    subs r2,r2,#1
  bne mYloop
 }
}



void draw_backdrop()
{
  __asm
  {
    mov r0,VRAM
    mov r1,backdrop
    mov r2,#0
    d_bdYloop:
      add r3,r1,r2
      ldrb r4,[r3]

      // r4 = color | (color<<8) | (color<<16) | (color<<24)
      mov r3,r4,lsl #8
      orr r3,r3,r4
      mov r4,r3,lsl #16
      orr r4,r4,r3

      mov r3,r4
      mov r5,#30	// 30*8 = 240 pixels
      d_bdXloop:
        // store 8 bytes per iteration
        stmia r0!,{r3-r4}
        subs r5,r5,#1
      bne d_bdXloop
      adds r2,r2,#1
      cmp r2,#100
    bne d_bdYloop
  }
}


void put_char(int x, int y, u8 col, u8 tecken)
{
u8 *src,*dest;
int xdim,ydim;

  src = (font+(char_size*tecken))+260;
  xdim = *(u8 *)(font+tecken+4);
  ydim = font_height;
  dest = (u8 *)(VRAM+(y*240)+x);
  __asm 
  {
    mov r0,src
    mov r1,dest
    mov r2,ydim
    mov r3,font_width
    mov r5,xdim
    sub r3,r3,r5
    mov r4,#240
    sub r4,r4,r5
    mov r7,col
    pcYloop:
      mov r5,xdim
      pcXloop:
        ldrb r6,[r0]
        orrs r6,r6,r6
        beq invisible
          strb r7,[r1]
        invisible:
        add r0,r0,#1
        add r1,r1,#1
        subs r5,r5,#1
      bne pcXloop
      add r0,r0,r3  
      add r1,r1,r4
      subs r2,r2,#1
    bne pcYloop
  }
}



 

 
int GBA_main()
{
int i, theta=0;
float f;
u8 a1=0,a2,col=0;
int xpos,ypos;


  font_width = *(u8 *)(font+2);
  font_height = *(u8 *)(font+3);
  char_size = (font_width*font_height);

  for (f=0; f<256; f+=1)
  {
    sine[(int)f] = (float)sin(f*3.1415926f/128.0f);
    cosine[(int)f] = (float)cos(f*3.1415926f/128.0f);
  }

  gbalSetMode(_240x160x1x256_);

  // Set up the palette.
  for (i=0; i<64; i++)
   {
    gbalSetColor(BACKGROUND, i, ((i&62)<<9));
    gbalSetColor(BACKGROUND, i+64, gbalMakeColor(90, 90, (((i<<2)+180)>>1)));
   }
  for (i=0; i<32; i++)
   {
    gbalSetColor(BACKGROUND, i+128, i);
    gbalSetColor(BACKGROUND, i+160, ((i<<5)|31));
    gbalSetColor(BACKGROUND, i+192, gbalMakeColor((((i<<3)+180)>>1), 90, 90));
    gbalSetColor(BACKGROUND, i+224, gbalMakeColor(214, (((i<<3)+180)>>1), 90));
   }


  triangle[0].c = 128;
  triangle[1].c = 191;
  triangle[2].c = 160;
  

  camX = 120;	camY = 50;	camZ = -300;

  gbalRContext((u8 *)VRAM,240,100,8);
  gbalShadeModel(GBAL_SMOOTH);


  while (1)
  {
    Rotate(&backup[0], &points[0], theta>>1, 0, theta, 3);
    Project(&points[0], &triangle[0].x, 3);

    gbalVSync();

    draw_backdrop();  
    gbalTriangle(&triangle[0], &triangle[1], &triangle[2]);

    a2 = a1;
    xpos = 88;
    for (i=0; i<9; i++)
    {
      ypos = (int)(sine[a2]*6)+80;
      put_char(xpos, ypos, col+150, message[i]);
      xpos += 8;
      col = (col+7)&31;
      a2 += 18;
    }
    a1++;
 
    mirror();
    theta++;

  }


  return 0;
}

       





